from typing import Type, Tuple

from wai.annotations.core.component import ProcessorComponent
from wai.annotations.core.domain import DomainSpecifier
from wai.annotations.core.specifier import ProcessorStageSpecifier


class CombineAnnotationsODISPSpecifier(ProcessorStageSpecifier):
    """
    Specifies the combine-annotations object detection ISP.
    """
    @classmethod
    def description(cls) -> str:
        return "Combines object detection annotations from images passing through into a single annotation."

    @classmethod
    def domain_transfer_function(
            cls,
            input_domain: Type[DomainSpecifier]
    ) -> Type[DomainSpecifier]:
        from wai.annotations.domain.image.object_detection import ImageObjectDetectionDomainSpecifier
        if input_domain is ImageObjectDetectionDomainSpecifier:
            return input_domain
        else:
            raise Exception(
                f"CombineAnnotationsOD only handles the following domains: "
                f"{ImageObjectDetectionDomainSpecifier.name()}"
            )

    @classmethod
    def components(cls) -> Tuple[Type[ProcessorComponent]]:
        from wai.annotations.imgvis.isp.combine_annotations.component import CombineAnnotationsOD
        return CombineAnnotationsOD,



from shapely.geometry import Polygon, GeometryCollection, MultiPolygon
from shapely.ops import unary_union

from wai.common.cli.options import TypedOption
from wai.common.geometry import Polygon as WaiPolygon
from wai.common.geometry import Point as WaiPoint
from wai.common.adams.imaging.locateobjects import LocatedObjects, LocatedObject
from wai.annotations.core.component import ProcessorComponent
from wai.annotations.core.stream import ThenFunction, DoneFunction
from wai.annotations.core.stream.util import RequiresNoFinalisation
from wai.annotations.domain.image.object_detection import ImageObjectDetectionInstance


UNION = "union"
INTERSECT = "intersect"
COMBINATIONS = [
    UNION,
    INTERSECT,
]

STREAM_INDEX = "stream_index"


def to_polygons(located_objects):
    """
    Turns the located objects into shapely polygons.

    :param located_objects: the objects to convert
    :type located_objects: LocatedObjects
    :return: the list of polygons
    :rtype: list
    """
    result = []
    for obj in located_objects:
        coords = []
        for point in obj.get_polygon().points:
            coords.append((point.x, point.y))
        result.append(Polygon(coords))
    return result


def intersect_over_union(poly1, poly2):
    """
    Calculates the IoU (intersect over union) for the two polygons.

    :param poly1: the first polygon
    :type poly1: Polygon
    :param poly2: the second polygon
    :type poly2: Polygon
    :return: the IoU
    :type: float
    """
    try:
        intersection = poly2.intersection(poly1)
        if intersection.area > 0:
                union = unary_union([poly2, poly1])
                return intersection.area / union.area
        else:
            return 0
    except:
        print("Failed to compute IoU!")
        return 0


class CombineAnnotationsOD(
    RequiresNoFinalisation,
    ProcessorComponent[ImageObjectDetectionInstance, ImageObjectDetectionInstance]
):
    """
    Stream processor which combines object detections into single one.
    """

    min_iou: float = TypedOption(
        "--min-iou",
        type=float,
        default=0.7,
        help="the minimum IoU (intersect over union) to use for identifying objects that overlap"
    )

    combination: str = TypedOption(
        "--combination",
        type=str,
        default=INTERSECT,
        help="how to combine the annotations (%s); the '%s' key in the meta-data contains the stream index" % ("|".join(COMBINATIONS), STREAM_INDEX)
    )

    def _find_matches(self, polygons_old, polygons_new):
        """
        Finds the matches between the old and new annotations.

        :param polygons_old: the old annotations
        :type polygons_old: list
        :param polygons_new: the new annotations
        :type polygons_new: list
        :return: the matches, list of old/new index tuples (an index of -1 means no match found)
        :rtype: list
        """
        result = []
        match_new = set([x for x in range(len(polygons_new))])
        match_old = set([x for x in range(len(polygons_old))])
        for n, poly_new in enumerate(polygons_new):
            for o, poly_old in enumerate(polygons_old):
                iou = intersect_over_union(poly_new, poly_old)
                if iou > 0:
                    if iou >= self.min_iou:
                        if n in match_new:
                            match_new.remove(n)
                        if o in match_old:
                            match_old.remove(o)
                        result.append((o, n, iou))

        # add old polygons that had no match
        for o in match_old:
            result.append((o, -1, 0.0))

        # add new polygons that had no match
        for n in match_new:
            result.append((-1, n, 0.0))

        return result

    def process_element(
            self,
            element: ImageObjectDetectionInstance,
            then: ThenFunction[ImageObjectDetectionInstance],
            done: DoneFunction
    ):
        if not hasattr(self, "_annotations"):
            self._annotations = element.annotations
            self._stream_index = 0
            then(element)
            return

        self._stream_index += 1

        # combine annotations
        polygons_old = to_polygons(self._annotations)
        polygons_new = to_polygons(element.annotations)
        matches = self._find_matches(polygons_old, polygons_new)
        combined = []
        for o, n, iou in matches:
            if o == -1:
                combined.append(element.annotations[n])
            elif n == -1:
                combined.append(self._annotations[o])
            else:
                # combine polygons
                if self.combination == UNION:
                    poly_comb = unary_union([polygons_new[n], polygons_old[o]])
                elif self.combination == INTERSECT:
                    poly_comb = polygons_new[n].intersection(polygons_old[o])
                else:
                    raise Exception("Unknown combination method: %s" % self.combination)
                # grab the first polygon
                if isinstance(poly_comb, GeometryCollection):
                    for x in poly_comb.geoms:
                        if isinstance(x, Polygon):
                            poly_comb = x
                            break
                elif isinstance(poly_comb, MultiPolygon):
                    for x in poly_comb.geoms:
                        if isinstance(x, Polygon):
                            poly_comb = x
                            break

                if isinstance(poly_comb, Polygon):
                    # create new located object
                    minx, miny, maxx, maxy = [int(x) for x in poly_comb.bounds]
                    x_list, y_list = poly_comb.exterior.coords.xy
                    points = []
                    for i in range(len(x_list)):
                        points.append(WaiPoint(x=x_list[i], y=y_list[i]))
                    lobj = LocatedObject(minx, miny, maxx - minx + 1, maxy - miny + 1)
                    lobj.set_polygon(WaiPolygon(*points))
                    lobj.metadata[STREAM_INDEX] = self._stream_index
                    combined.append(lobj)
                else:
                    self.logger.warning("Unhandled geometry type returned from combination, skipping: %s" % str(type(poly_comb)))

        self._annotations = LocatedObjects(combined)

        # new element
        then(element.__class__(element.data, self._annotations))



import cv2
import io
import numpy as np

from wai.common.cli.options import TypedOption
from wai.annotations.core.component import SinkComponent
from wai.annotations.domain.image import ImageInstance


class ImageViewer(
    SinkComponent[ImageInstance]
):
    """
    Sink for displaying images.
    """

    title: str = TypedOption(
        "--title",
        type=str,
        default="wai.annotations",
        help="the title for the window"
    )

    position: str = TypedOption(
        "--position",
        type=str,
        default="0,0",
        help="the position of the window on screen (X,Y)"
    )

    size: str = TypedOption(
        "--size",
        type=str,
        default="640,480",
        help="the maximum size for the image: WIDTH,HEIGHT"
    )

    delay: int = TypedOption(
        "--delay",
        type=int,
        default=500,
        help="the delay in milli-seconds between images, use 0 to wait for keypress, ignored if <0"
    )

    def consume_element(self, element: ImageInstance):
        """
        Consumes instances by displaying them.
        """
        # read image
        img_array = np.fromstring(io.BytesIO(element.data.data).read(), dtype=np.uint8)
        img = cv2.imdecode(img_array, cv2.IMREAD_COLOR)

        # resize image, if necessary
        if not hasattr(self, "_width"):
            self._width, self._height = [int(x) for x in self.size.split(",")]
            self._ratio = self._width / self._height
        h, w, _ = img.shape
        if (h > self._height) or (w > self._width):
            img_ratio = w / h
            if img_ratio > self._ratio:
                w_new = self._width
                h_new = w_new / img_ratio
            else:
                h_new = self._height
                w_new = h_new * img_ratio
            img = cv2.resize(img, (int(w_new), int(h_new)))

        cv2.imshow(self.title, img)

        # position window
        if not hasattr(self, "_x"):
            self._x, self._y = [int(x) for x in self.position.split(",")]
            cv2.moveWindow(self.title, self._x, self._y)

        # delay
        if self.delay >= 0:
            cv2.waitKey(self.delay)

    def finish(self):
        cv2.destroyAllWindows()



import io
import PIL

from typing import List
from PIL import ImageDraw

from wai.common.cli.options import TypedOption
from wai.annotations.core.component import ProcessorComponent
from wai.annotations.core.stream import ThenFunction, DoneFunction
from wai.annotations.core.stream.util import RequiresNoFinalisation
from wai.annotations.domain.image import Image
from wai.annotations.domain.image.segmentation import ImageSegmentationInstance
from wai.annotations.imgvis.isp.annotation_overlay.component._colors import default_colors


class AnnotationOverlayIS(
    RequiresNoFinalisation,
    ProcessorComponent[ImageSegmentationInstance, ImageSegmentationInstance]
):
    """
    Stream processor which adds image classification labels on top of images.
    """

    labels: List[str] = TypedOption(
        "--labels",
        type=str,
        nargs="+",
        help="the labels of annotations to overlay, overlays all if omitted"
    )

    alpha: int = TypedOption(
        "--alpha",
        type=int,
        default=64,
        help="the alpha value to use for overlaying the annotations (0: transparent, 255: opaque)."
    )

    colors: List[str] = TypedOption(
        "--colors",
        type=str,
        nargs="+",
        help="the RGB triplets (R,G,B) of custom colors to use, uses default colors if not supplied"
    )

    def _initialize(self):
        """
        Initializes colors etc.
        """
        self._colors = dict()
        self._default_colors = default_colors()
        self._default_colors_index = 0
        self._custom_colors = []
        if self.colors is not None:
            for color in self.colors:
                self._custom_colors.append([int(x) for x in color.split(",")])
        self._accepted_labels = None
        if (self.labels is not None) and (len(self.labels) > 0):
            self._accepted_labels = set(self.labels)

    def _next_default_color(self):
        """
        Returns the next default color.

        :return: the color tuple
        :rtype: tuple
        """
        if self._default_colors_index >= len(self._default_colors):
            self._default_colors_index = 0
        result = self._default_colors[self._default_colors_index]
        self._default_colors_index += 1
        return result

    def _get_color(self, label):
        """
        Returns the color for the label.

        :param label: the label to get the color for
        :type label: str
        :return: the RGBA color tuple
        :rtype: tuple
        """
        if label not in self._colors:
            has_custom = False
            if label in self._label_mapping:
                index = self._label_mapping[label]
                if index < len(self._custom_colors):
                    has_custom = True
                    self._colors[label] = self._custom_colors[index]
            if not has_custom:
                self._colors[label] = self._next_default_color()
        r, g, b = self._colors[label]
        return r, g, b, self.alpha

    def process_element(
            self,
            element: ImageSegmentationInstance,
            then: ThenFunction[ImageSegmentationInstance],
            done: DoneFunction
    ):
        if not hasattr(self, "_colors"):
            self._initialize()

        img_in = element.data
        img_pil = element.data.pil_image

        # create label/index mapping for custom colors
        self._label_mapping = dict()
        for index, label in enumerate(element.annotations.labels):
            self._label_mapping[label] = index

        # create overlay for annotations
        overlay = PIL.Image.new('RGBA', img_pil.size, (0, 0, 0, 0))
        draw = ImageDraw.Draw(overlay)

        label_images = element.annotations.label_images
        updated = False
        for label in label_images:
            # skip label?
            if (self._accepted_labels is not None) and (label not in self._accepted_labels):
                continue
            # draw overlay
            updated = True
            mask = label_images[label]
            draw.bitmap((0, 0), mask, fill=self._get_color(label))

        if updated:
            # add overlay
            img_pil.paste(overlay, (0, 0), mask=overlay)
            # convert back to PIL bytes
            pil_img_bytes = io.BytesIO()
            img_pil.save(pil_img_bytes, format=img_in.format.pil_format_string)
            img_out = Image(img_in.filename, pil_img_bytes.getvalue(), img_in.format, img_in.size)

            # new element
            then(element.__class__(img_out, element.annotations))
        else:
            then(element)



import io
import PIL

from typing import List
from PIL import ImageDraw

from wai.common.cli.options import TypedOption, FlagOption
from wai.annotations.core.component import ProcessorComponent
from wai.annotations.core.stream import ThenFunction, DoneFunction
from wai.annotations.core.stream.util import RequiresNoFinalisation
from wai.annotations.domain.image import Image
from wai.annotations.domain.image.object_detection import ImageObjectDetectionInstance
from wai.annotations.imgvis.isp.annotation_overlay.component._colors import default_colors, text_color
from wai.annotations.imgvis.isp.annotation_overlay.component._fonts import DEFAULT_FONT_FAMILY, load_font


class AnnotationOverlayOD(
    RequiresNoFinalisation,
    ProcessorComponent[ImageObjectDetectionInstance, ImageObjectDetectionInstance]
):
    """
    Stream processor which adds object detections overlays to images.
    """

    labels: List[str] = TypedOption(
        "--labels",
        type=str,
        nargs="+",
        help="the labels of annotations to overlay, overlays all if omitted"
    )

    label_key: str = TypedOption(
        "--label-key",
        type=str,
        default="type",
        help="the key in the meta-data that contains the label."
    )

    text_format: str = TypedOption(
        "--text-format",
        type=str,
        default="{label}",
        help="template for the text to print on top of the bounding box or polygon, '{PH}' is a placeholder for the 'PH' value from the meta-data or 'label' for the current label; ignored if empty."
    )

    text_placement: str = TypedOption(
        "--text-placement",
        type=str,
        default="T,L",
        help="comma-separated list of vertical (T=top, C=center, B=bottom) and horizontal (L=left, C=center, R=right) anchoring."
    )

    font_family: str = TypedOption(
        "--font-family",
        type=str,
        default=DEFAULT_FONT_FAMILY,
        help="the name of the TTF font-family to use, note: any hyphens need escaping with backslash."
    )

    font_size: int = TypedOption(
        "--font-size",
        type=int,
        default=14,
        help="the size of the font."
    )

    num_decimals: int = TypedOption(
        "--num-decimals",
        type=int,
        default=3,
        help="the number of decimals to use for float numbers in the text format string."
    )

    colors: List[str] = TypedOption(
        "--colors",
        type=str,
        nargs="+",
        help="the RGB triplets (R,G,B) of custom colors to use, uses default colors if not supplied"
    )

    outline_thickness: int = TypedOption(
        "--outline-thickness",
        type=int,
        default=3,
        help="the line thickness to use for the outline, <1 to turn off."
    )

    outline_alpha: int = TypedOption(
        "--outline-alpha",
        type=int,
        default=255,
        help="the alpha value to use for the outline (0: transparent, 255: opaque)."
    )

    fill: bool = FlagOption(
        "--fill",
        help="whether to fill the bounding boxes/polygons"
    )

    fill_alpha: int = TypedOption(
        "--fill-alpha",
        type=int,
        default=128,
        help="the alpha value to use for the filling (0: transparent, 255: opaque)."
    )

    vary_colors: bool = FlagOption(
        "--vary-colors",
        help="whether to vary the colors of the outline/filling regardless of label"
    )

    force_bbox: bool = FlagOption(
        "--force-bbox",
        help="whether to force a bounding box even if there is a polygon available"
    )

    def _initialize(self):
        """
        Initializes colors etc.
        """
        self._colors = dict()
        self._default_colors = default_colors()
        self._default_colors_index = 0
        self._custom_colors = []
        if self.colors is not None:
            for color in self.colors:
                self._custom_colors.append([int(x) for x in color.split(",")])
        self._label_mapping = dict()
        self._font = load_font(self.logger, self.font_family, self.font_size)
        self._text_vertical, self._text_horizontal = self.text_placement.upper().split(",")
        self._accepted_labels = None
        if (self.labels is not None) and (len(self.labels) > 0):
            self._accepted_labels = set(self.labels)

    def _next_default_color(self):
        """
        Returns the next default color.

        :return: the color tuple
        :rtype: tuple
        """
        if self._default_colors_index >= len(self._default_colors):
            self._default_colors_index = 0
        result = self._default_colors[self._default_colors_index]
        self._default_colors_index += 1
        return result

    def _get_color(self, label):
        """
        Returns the color for the label.

        :param label: the label to get the color for
        :type label: str
        :return: the RGB color tuple
        :rtype: tuple
        """
        if label not in self._colors:
            has_custom = False
            if label in self._label_mapping:
                index = self._label_mapping[label]
                if index < len(self._custom_colors):
                    has_custom = True
                    self._colors[label] = self._custom_colors[index]
            if not has_custom:
                self._colors[label] = self._next_default_color()
        return self._colors[label]

    def _get_outline_color(self, label):
        """
        Generates the color for the outline.

        :param label: the label to get the color for
        :type label: str
        :return: the RGBA color tuple
        :rtype: tuple
        """
        r, g, b = self._get_color(label)
        return r, g, b, self.outline_alpha

    def _get_fill_color(self, label):
        """
        Generates the color for the filling.

        :param label: the label to get the color for
        :type label: str
        :return: the RGBA color tuple
        :rtype: tuple
        """
        r, g, b = self._get_color(label)
        return r, g, b, self.fill_alpha

    def _expand_label(self, label, metadata):
        """
        Expands the label text.

        :param label: the current label
        :type label: str
        :param metadata: the metadata associated with the label
        :type metadata: dict
        :return: the expanded label text
        :rtype: str
        """
        result = self.text_format.replace("{label}", label)
        for key in metadata:
            value = metadata[key]
            if isinstance(value, str) or isinstance(value, int) or isinstance(value, bool):
                result = result.replace("{%s}" % key, str(value))
            elif isinstance(value, float):
                result = result.replace("{%s}" % key, ("%." + str(self.num_decimals) + "f") % float(value))
        return result

    def _text_coords(self, draw, text, rect):
        """
        Determines the text coordinates in the image.

        :param draw: the ImageDraw instance
        :type draw: ImageDraw
        :param text: the text to output
        :type text: str
        :param rect: the rectangle to use as reference
        :return: the x, y, w, h tuple
        :rtype: tuple
        """
        w, h = draw.textsize(text, font=self._font)

        # x
        if self._text_horizontal == "L":
            x = rect.left()
        elif self._text_horizontal == "C":
            x = rect.left() + (rect.right() - rect.left() - w) // 2
        elif self._text_horizontal == "R":
            x = rect.right() - w
        else:
            raise Exception("Unhandled horizontal text position: %s" % self._text_horizontal)

        # y
        if self._text_vertical == "T":
            y = rect.top()
        elif self._text_vertical == "C":
            y = rect.top() + (rect.bottom() - rect.top() - h) // 2
        elif self._text_vertical == "B":
            y = rect.bottom() - h
        else:
            raise Exception("Unhandled horizontal text position: %s" % self._text_horizontal)

        return x, y, w, h

    def process_element(
            self,
            element: ImageObjectDetectionInstance,
            then: ThenFunction[ImageObjectDetectionInstance],
            done: DoneFunction
    ):
        if not hasattr(self, "_colors"):
            self._initialize()

        img_in = element.data
        img_pil = element.data.pil_image

        overlay = PIL.Image.new('RGBA', img_pil.size, (0, 0, 0, 0))
        draw = ImageDraw.Draw(overlay)
        for i, lobj in enumerate(element.annotations):
            # determine label/color
            label = "object"
            if self.label_key in lobj.metadata:
                label = lobj.metadata[self.label_key]
            if self._accepted_labels is not None:
                if label not in self._accepted_labels:
                    continue
            if label not in self._label_mapping:
                self._label_mapping[label] = len(self._label_mapping)
            if self.vary_colors:
                color_label = "object-%d" % i
            else:
                color_label = label

            # assemble polygon
            points = []
            if lobj.has_polygon() and not self.force_bbox:
                poly_x = lobj.get_polygon_x()
                poly_y = lobj.get_polygon_y()
                for x, y in zip(poly_x, poly_y):
                    points.append((x, y))
            else:
                rect = lobj.get_rectangle()
                points.append((rect.left(), rect.top()))
                points.append((rect.right(), rect.top()))
                points.append((rect.right(), rect.bottom()))
                points.append((rect.left(), rect.bottom()))
            if self.fill:
                draw.polygon(tuple(points), outline=self._get_outline_color(color_label), fill=self._get_fill_color(color_label), width=self.outline_thickness)
            else:
                draw.polygon(tuple(points), outline=self._get_outline_color(color_label), width=self.outline_thickness)

            # output text
            if len(self.text_format) > 0:
                text = self._expand_label(label, lobj.metadata)
                rect = lobj.get_rectangle()
                x, y, w, h = self._text_coords(draw, text, rect)
                draw.rectangle((x, y, x+w, y+h), fill=self._get_outline_color(color_label))
                draw.text((x, y), text, font=self._font, fill=text_color(self._get_color(color_label)))

        img_pil.paste(overlay, (0, 0), mask=overlay)

        # convert back to PIL bytes
        pil_img_bytes = io.BytesIO()
        img_pil.save(pil_img_bytes, format=img_in.format.pil_format_string)
        img_out = Image(img_in.filename, pil_img_bytes.getvalue(), img_in.format, img_in.size)

        # new element
        then(element.__class__(img_out, element.annotations))
